#include <iostream>
#include <fstream>
#include <sstream>
#include <list>
#include <map>
#include <set>
#include <vector>
#include <string>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

struct Item
{
    void print() const;

    int m_id;
    std::string m_mapid;
    std::string m_key;
    std::string m_name;   
    std::string m_slot;
    int m_owner;
    int m_unit;
    int m_position[3];
    int m_ammoQty;
    int m_ammoItem;
     
    Item* m_ammo;
};

typedef std::map<int, Item *> ItemMap;
typedef std::list<Item *> ItemList;

struct Unit
{
    void print(bool live) const;
    void mapPrint(bool showPiles) const;

    int m_id;
    std::string m_key;   
    std::string m_name;   
    std::string m_armorName;
    char m_symbol;
    int m_hp;
    int m_stun;
    int m_morale;
    int m_tu;
    int m_position[3];
    int m_armor[5];
    int m_wounds;
    int m_faction;
    int m_itemcount;
    bool m_surrender;
    char m_rank;

    ItemList m_items;
};

typedef std::map<int, Unit *> UnitMap;

struct Naming
{
    std::string m_name;
    std::string m_pedia;
};

typedef std::map<std::string, Naming*> NameMap;

//  Parse
//  Find battleGame:
//  Find '  units:'
//  Each '    -id:'  is a unit, add to unit map.
//  For each unit, get genUnitType,
//      position,health,stunlevel,morale,wantsToSurrender
//  armor:  values, total fatalWounds:
//  Find '  items:'
//  Each '    id:'  is an item.
//     Get type, owner, position, ammoItem. ammoqty
//  Items key in map by id.
//  Each name key, add to translation map.
//  Translate: Read lang file, get names for all string keys.
//  Correlate:
//  Any item with ammoItem - look up and link.
//  Any item with owner - add to that unit id list.
//  Any item on ground - add to ground pile at those coords
//  Result: show live enemy locations and stats, what they have, and loaded with what.
//  Show knocked out enemy locations/stats.
//  Show ground item piles.  

struct Battle
{
    Battle(const std::string& saveFileName);

    void translate(const std::string& langFileName);
    void processItems();
    void print(bool showPiles, bool showMap, int level);

    ItemMap m_items;
    UnitMap m_units;
    NameMap m_names;

    int m_mapWidth;
    int m_mapLength;
};

bool hasValue(const std::string& line, const char* tag, std::string& val)
{
    size_t pos = line.find(tag);
    if(pos == std::string::npos) return false;
    pos += strlen(tag);
    pos = line.find_first_not_of(' ', pos);
    if(pos != std::string::npos) val = line.substr(pos);
    return true;
}

int itemQty(const std::string& line, std::string& val)
{
    size_t pos = line.find("STR_");
    if(pos == std::string::npos) return 0;
    size_t colpos = line.find(':', pos);
    if(colpos == std::string::npos) return 0;
    val = line.substr(pos, colpos - pos);
    return atoi(line.substr(colpos + 2).c_str());
}

void trimTrailing(std::string& line)
{
    size_t pos = line.find_last_not_of("\n\r\t ");
    if(pos != std::string::npos && pos < line.size())
    {
        line.erase(pos+1);
    }
}

int readArray(std::ifstream& handle, int n, int* arr)
{
    std::string sLine;
    std::string val;
    int i;
    for(i = 0; i < n; ++i)
    {
        if(!getline(handle, sLine)) break;
        trimTrailing(sLine);
        if(!hasValue(sLine, "- ", val)) break;
        arr[i] = atoi(val.c_str());
    }
    return i;
}

void Item::print() const
{
    if(!m_ammo)
    {
        std::cout << "    " << m_name << std::endl;
    }
    else
    {
        std::cout << "    " << m_name << " ( " << m_ammo->m_name << ", " << m_ammo->m_ammoQty << " )" << std::endl;
    }
}

// live = true means show living units (including passed out)
// live = false means show dead units or item piles
void Unit::print(bool live) const
{
    std::string status;
    std::stringstream oss;
    if(m_hp > 0)
    {
        if(!live) return;        
        oss << " [ ";
        if(m_wounds > 0 || m_stun >= m_hp || m_surrender)
        {
            if(m_wounds > 0) oss << "BLEEDING ";
            if(m_stun >= m_hp) oss << "STUNNED ";
            if(m_surrender) oss << "SURRENDER ";
        }
        else oss << "OK ";
        oss << "]";
        status = oss.str();
    }
    else
    {
        if(live) return;
        if(m_faction >= 0)
        {
            status.assign(" [ DEAD ]");
        }
        else
        {
            oss << " [ " << m_itemcount << " ITEMS ]";
            status = oss.str();
        }
    }

    if(m_id < 1000000) std::cout << "S" << m_symbol << ": ";
    else if(m_id < 2000000) std::cout << (m_id - 1000000) << ": ";
    else std::cout << "Pile: ";

    std::cout << m_name << status << ", at [ " 
              << m_position[0] << ", " << m_position[1] << ", " << m_position[2] << " ]" << std::endl;

    if(m_hp > 0)
    {
        std::cout << "  HP: " << m_hp << ", Stun: " << m_stun << ", Morale: " << m_morale 
                  << ", Wounds: " << m_wounds << ", TU: " << m_tu << std::endl;
        std::cout << "  Armor [ " 
                  << m_armor[0] << " " 
                  << m_armor[1] << " " 
                  << m_armor[2] << " " 
                  << m_armor[3] << " "
                  << m_armor[4] << " ]" << std::endl;
    }
    
    if(m_itemcount > 0 && m_itemcount <= 50)
    {
        std::cout << "  ITEMS:" << std::endl;
        ItemList::const_iterator iit = m_items.begin();
        while(iit != m_items.end())
        {
            (*iit)->print();
            ++iit;
        }
    }
    std::cout << std::endl;
}

void Unit::mapPrint(bool showPiles) const
{
    if(m_id >= 2000000 || m_hp <= 0) // item pile or dead
    {
        if(!showPiles) std::cout << ". ";
        else std::cout << "XX";
    }
    else if(m_id >= 1000000) // live enemy unit
    {
        char buf[10];
        sprintf(buf, "%02d", m_id - 1000000);
        std::cout << buf;
    }
    else // soldier
    {
        std::cout << 'S' << m_symbol;
    }
}

Battle::Battle(const std::string& saveFileName) 
{
    std::ifstream saveFile(saveFileName.c_str());
    if (!saveFile)
    {
        std::cerr << "Unable to open " << saveFileName << std::endl;
        return;
    }

    std::string sLine;
    std::string val;    

    // Run through bases, extract soldier IDs/names
    typedef std::map<int, std::string> SoldierMap;    
    SoldierMap soldierMap;
    bool haveSoldierList = false;
    bool haveSoldier = false;
    int soldierId;

    while ( getline(saveFile, sLine))
    {
        trimTrailing(sLine);
        if(sLine.find("battleGame:") == 0) break;        
        if(!haveSoldierList && sLine.find("    soldiers:") == 0) 
        {
            haveSoldierList = true;
            haveSoldier = false;
            continue;
        }
        if(haveSoldierList)
        {
            if(sLine.find_first_not_of(' ') <= 4) 
            {
                haveSoldierList = false;
                continue;
            }  
            if(!haveSoldier && hasValue(sLine, "- type:", val))
            {
                haveSoldier = true;
                continue;
            }
            if(haveSoldier)
            {
                if(hasValue(sLine, "id:", val))
                {
                    soldierId = atoi(val.c_str());                    
                }
                else if(hasValue(sLine, "name:", val))
                {
                    soldierMap[soldierId] = val;
                    haveSoldier = false;
                }
            }
        }
    }

    char mode = '\0';

    Item* item = NULL;
    Unit* unit = NULL;
    
    while ( getline(saveFile, sLine))
    {
        trimTrailing(sLine);
        if(sLine[0] != ' ') break;
        if(sLine[2] != ' ')
        {
            if(sLine.find("  units:") == 0) mode = 'U';
            else if(sLine.find("  items:") == 0) mode = 'I';
            else if(hasValue(sLine, "width:", val))
            {
                m_mapWidth = atoi(val.c_str());
            }
            else if(hasValue(sLine, "length:", val))
            {
                m_mapLength = atoi(val.c_str());
            }      
            else mode = '\0';
            continue;
        }
        if(mode == 'U')
        {
            if(hasValue(sLine, "- id:", val))
            {
                unit = new Unit();
                int id = atoi(val.c_str());
                m_units[id] = unit;
                unit->m_id = id;
                if(id < 1000000)
                {
                    SoldierMap::iterator it = soldierMap.find(id);
                    if(it != soldierMap.end()) unit->m_name = it->second;
                }
            }
            else if(unit)
            {
                if(hasValue(sLine, "genUnitType:", val))
                {
                    unit->m_key = val;
                    m_names[val] = NULL;
                }    
                else if(hasValue(sLine, "genUnitArmor:", val))     
                {
                    unit->m_armorName = val;
                }
                else if(hasValue(sLine, "wantsToSurrender:", val))
                {
                    unit->m_surrender = (val == "true");
                }
                else if(sLine.find("   position:") != std::string::npos)
                {
                    readArray(saveFile, 3, &(unit->m_position[0]));
                }
                else if(hasValue(sLine, "health:", val))
                {
                    unit->m_hp = atoi(val.c_str());
                }
                else if(hasValue(sLine, "tu:", val))
                {
                    unit->m_tu = atoi(val.c_str());
                }
                else if(hasValue(sLine, "stunlevel:", val))
                {
                    unit->m_stun = atoi(val.c_str());
                }
                else if(hasValue(sLine, "morale:", val))
                {
                    unit->m_morale = atoi(val.c_str());
                }
                else if(hasValue(sLine, "faction:", val))
                {
                    unit->m_faction = atoi(val.c_str());
                }
                else if(hasValue(sLine, "rankInt:", val))
                {
                    unit->m_rank = val[0];
                } 
                else if(sLine.find("   armor:") != std::string::npos)
                {
                    readArray(saveFile, 5, &(unit->m_armor[0]));
                }
                else if(sLine.find("   fatalWounds:") != std::string::npos)
                {
                    int fw[5];
                    readArray(saveFile, 5, &(fw[0]));
                    unit->m_wounds = 0;
                    for(int i = 0; i < 5; ++i) unit->m_wounds += fw[i];
                }
            }
        }
        else if(mode == 'I')
        {
            if(hasValue(sLine, "- id:", val))
            {
                item = new Item();
                int id = atoi(val.c_str());
                m_items[id] = item;
                item->m_id = id;
            }
            else if(item)
            {
                if(hasValue(sLine, "type:", val))
                {
                    item->m_key = val;
                    m_names[val] = NULL;
                } 
                else if(hasValue(sLine, "owner:", val))
                {
                    item->m_owner = atoi(val.c_str());
                }
                else if(hasValue(sLine, "inventoryslot:", val))
                {
                    item->m_slot = val;
                } 
                else if(hasValue(sLine, "unit:", val))
                {
                    item->m_unit = atoi(val.c_str());
                }
                else if(sLine.find("   position:") != std::string::npos)
                {
                    readArray(saveFile, 3, &(item->m_position[0]));
                }
                else if(hasValue(sLine, "ammoqty:", val))
                {
                    item->m_ammoQty = atoi(val.c_str());
                }
                else if(hasValue(sLine, "ammoItem:", val))
                {
                    item->m_ammoItem = atoi(val.c_str());
                }
            }            
        }
    }
}

void Battle::translate(const std::string& langFileName)
{
    std::ifstream langFile(langFileName.c_str());
    if (!langFile)
    {
        std::cerr << "Unable to open " << langFileName << std::endl;
        return;
    }

    std::string sLine;
    while ( getline(langFile, sLine))
    {
        trimTrailing(sLine);
        size_t pos = sLine.find("STR_");
        if(pos == std::string::npos) continue;
        size_t colpos = sLine.find(':', pos);
        if(colpos == std::string::npos) continue;
        size_t q1pos = sLine.find_first_not_of(": ", colpos);
        if(q1pos == std::string::npos) continue;
        size_t q2pos;
        if(sLine[q1pos] == '"') // Value is in quotes
        {
            ++q1pos;
            q2pos = sLine.find_first_of('"', q1pos);
            if(q2pos == std::string::npos) continue;
        }
        else // No quotes
        {
            q2pos = sLine.find_last_not_of(" \n\r\t");
            if(q2pos < q1pos) continue;
            q2pos++;
        }
        std::string key = sLine.substr(pos, colpos - pos);
        std::string val = sLine.substr(q1pos, q2pos - q1pos);
                // Is it an ufopedia entry?
        size_t upos = key.find("_UC_UFOPEDIA");      
        if(upos == std::string::npos) upos = key.find("_UFOPEDIA"); 
       
        NameMap::iterator it ;
        if(upos != std::string::npos)
        {
            it = m_names.find(key.substr(0, upos));
        }
        else
        {
            it = m_names.find(key);
        }    
        if(it != m_names.end())
        {
            if(!it->second) it->second = new Naming();        
            if(upos == std::string::npos) it->second->m_name = val;
            else it->second->m_pedia = val;
        }
    }

    // Apply translations
    UnitMap::iterator it = m_units.begin();
    while(it != m_units.end())
    {        
        if(it->second->m_name.empty())
        {
            NameMap::iterator nit = m_names.find(it->second->m_key);
            if(nit != m_names.end() && nit->second && !nit->second->m_name.empty())
            {
                it->second->m_name = nit->second->m_name;
            }
            else it->second->m_name = it->second->m_key;
        }
        ++it;
    }

    ItemMap::iterator iit = m_items.begin();
    while(iit != m_items.end())
    {
        NameMap::iterator nit = m_names.find(iit->second->m_key);
        if(nit != m_names.end() && nit->second && !nit->second->m_name.empty())
        {
            iit->second->m_name = nit->second->m_name;
        }
        else iit->second->m_name = iit->second->m_key;
        ++iit;
    }
}
void Battle::processItems()
{
    std::list<Unit*> piles;
    UnitMap::iterator it = m_units.begin();    
    char symbol = 'A';
    while(it != m_units.end())
    {
        Unit*unit = it->second;        
        unit->m_itemcount = 0;
        ++it;
        
        if(unit->m_id < 1000000) 
        {
            unit->m_symbol = symbol;
            ++symbol;
        }

        // If unit is dead, or passed out, add it to list of piles
        if(unit->m_hp <= 0 || unit->m_stun >= unit->m_hp)
        {
            piles.push_back(unit);
        }
    }   

    while(!piles.empty())
    {
        Unit* unit = piles.front();
        piles.pop_front();
        int poskey = unit->m_position[0] * 10000 + unit->m_position[1] * 100 + unit->m_position[2] + 2000000;
        m_units[poskey] = unit;
    }

    ItemMap::iterator iit = m_items.begin();
    while(iit != m_items.end())
    { 
        Item* item = iit->second;
        Unit* unit = NULL;
        ++iit;
        // Link to ammo
        item->m_ammo = NULL;
        if(item->m_ammoItem >= 0 && item->m_ammoItem != item->m_id)
        {
            ItemMap::iterator ait = m_items.find(item->m_ammoItem);
            if(ait != m_items.end()) item->m_ammo = ait->second;            
        }
        // Item represents a unit corpse?
        if(item->m_unit >= 0) continue;
        if(item->m_key.find("INV_NULL") != std::string::npos) continue;
        // Item owned by someone?
        if(item->m_owner > 0)
        {
            UnitMap::iterator uit = m_units.find(item->m_owner);
            if(uit != m_units.end())
            {
                uit->second->m_items.push_back(item);
                uit->second->m_itemcount++;
            }
            continue;
        }
        // Item on ground someplace.
        if(item->m_position[0] >= 0)
        {
            // Check if this pile already exists
            int poskey = item->m_position[0] * 10000 + item->m_position[1] * 100 + item->m_position[2] + 2000000;
            UnitMap::iterator uit = m_units.find(poskey);
            Unit* unit = NULL;
            if(uit != m_units.end()) 
                unit = uit->second;
            else
            {
                unit = new Unit();
                m_units[poskey] = unit;
                unit->m_id = poskey;
                unit->m_key.assign("GROUND_PILE");
                unit->m_name.assign("GROUND PILE");
                unit->m_position[0] = item->m_position[0];
                unit->m_position[1] = item->m_position[1];
                unit->m_position[2] = item->m_position[2];
                unit->m_faction = -1;
                unit->m_rank = ' ';
                unit->m_hp = 0;
                unit->m_itemcount = 0;
            }
            unit->m_items.push_back(item);
            unit->m_itemcount++;
            continue;
        }        
    } 
}

void Battle::print(bool showPiles, bool showMap, int level)
{   
    // Initialize map
    std::vector<Unit*> bmap;
    if(showMap)
    {
        bmap.reserve(m_mapLength * m_mapWidth);
        for(int i = 0; i < m_mapLength; ++i)
        {
            for(int j = 0; j < m_mapWidth; ++j)
            {
                bmap[i * m_mapWidth + j] = NULL;
            }
        }
    }

    UnitMap::iterator it = m_units.begin();
    while(it != m_units.end())
    {   
        Unit* unit = it->second;
        if(level < 0 || level == unit->m_position[2])
        {
            if(showPiles || (it->first < 2000000))
            {
                unit->print(it->first < 2000000);
                if(showMap)
                {
                    int pos = unit->m_position[0] * m_mapWidth + unit->m_position[1];
                    // If showing multiple levels, higher unit takes precedence
                    if(!bmap[pos] || bmap[pos]->m_position[2] < unit->m_position[2]) bmap[pos] = unit;
                }
            }
        }
        ++it;
    }  

    if(showMap)
    {
        std::cout << std::endl;

        for (int j = 0; j < m_mapWidth; ++j) 
        {
            for(int i = 0; i < m_mapLength; ++i) 
            {
                Unit* unit = bmap[i * m_mapWidth + j];
                if(!unit) std::cout << ". ";
                else unit->mapPrint(showPiles);
            }
            std::cout << std::endl;
        }
        std::cout << std::endl;
    }
}

int main(int argc, char **argv)   
{
    // Directories - Change Accordingly
    std::string rulesetDir("./");
    std::string savegameDir("./"); 
   
    // Defaut file names
    std::string langFileName("Piratez_lang.rul");
    std::string ruleFileName("Piratez.rul");
    std::string saveFileName("m");    

    bool showMap = false;
    bool showPiles = false;
    int level = -1;

    char option = '\0';

    for(int i = 1; i < argc; ++i)
    {
        if(argv[i][0] == '-') // option
        {
            option = argv[i][1];
            if(option == 'm')
            {
                showMap = true;
            }
            if(option == 'p')
            {
                showPiles = true;
            }
            continue;
        }
        switch(option)
        {
        case 'r': // rule file name
            ruleFileName.assign(argv[i]);
            break;
        case 'l': // lang file name
            langFileName.assign(argv[i]);
            break;
        case 's': // save file name
            saveFileName.assign(argv[i]);
            break;
        case 'm': // save file name
            level = atoi(argv[i]);
            break;
        default:
            break;
        }
        option = '\0';
    }

    // Parse save file
    std::string saveFile(savegameDir);    
    saveFile += saveFileName;
    saveFile += ".sav";
    
    Battle battle(saveFile);
    std::string langFile(rulesetDir);
    langFile += langFileName;    
    battle.translate(langFile);
    battle.processItems();
    battle.print(showPiles, showMap, level);
}
